import datetime
from data_tools.utils.times import parse_iso_datetime, ensure_utc
from typing import Union, List


type DateLike = Union[str, datetime.datetime]


class Event:
    """
    Represents an event that took place between `start` and `stop` (accessible as datetime.datetime objects or
    as ISO 8601 formatted strings), with a `name` and optionally any additional `attributes` such as
    `attributes["realtime"]` or `attributes["test"]`.
    """
    def __init__(
            self,
            start: DateLike,
            stop: DateLike, name: str = None,
            attributes: dict = None,
            flags: List[str] = None
    ):
        """
        Create an Event from either ISO8601 strings or datetime.datetime objects.
        Any datetime.datetime objects or ISO8601 strings MUST contain timezone information.

        Optionally, name the event with `name`, and with additional `attributes`.

        :param start: The start time of this Event
        :param stop: The end time of this Event
        :param name: The name of this event, optional, defaulted to "Unnamed Event"
        :param flags: A list of strings that indicate flags that are present.
        :param dict attributes: A dictionary of any additional attributes and/or metadata for this Event
        """
        if isinstance(start, datetime.datetime):
            self._start = ensure_utc(start)
        elif isinstance(start, str):
            self._start: datetime = parse_iso_datetime(start)
        else:
            raise TypeError("start must be datetime or str!")

        if isinstance(stop, datetime.datetime):
            self._stop = ensure_utc(stop)
        elif isinstance(stop, str):
            self._stop: datetime = parse_iso_datetime(stop)
        else:
            raise TypeError("start must be datetime or str!")

        if not isinstance(name, str) and name is not None:
            raise TypeError("name must be str!")
        else:
            self._name = name if name is not None else "Unnamed Event"

        if not isinstance(attributes, dict) and attributes is not None:
            raise TypeError("attributes must be dict!")
        else:
            self._attributes = attributes if attributes is not None else None

        if not isinstance(flags, list) and flags is not None:
            raise TypeError("flags must be list!")
        else:
            self._flags = flags if flags is not None else None

    @property
    def attributes(self) -> dict:
        """
        Obtain a dictionary of additional metadata or relevant attributes for this Event.
                """
        return self._attributes

    @property
    def start(self) -> datetime.datetime:
        """
        Obtain the end time of this Event as a datetime.datetime object.
        """
        return self._start

    @property
    def flags(self) -> List[str]:
        """
        Obtain the flags of this event as a list. May be empty.
        """
        return self._flags

    @property
    def stop(self) -> datetime.datetime:
        """
        Obtain the end time of this Event as a datetime.datetime object.
        """
        return self._stop

    @property
    def start_as_iso_str(self) -> str:
        """
        Return the start time of this Event as an ISO8601 string, such as 2024-11-07T15:30:45.12Z
        """
        return self._start.isoformat().replace("+00:00", "Z")

    @property
    def stop_as_iso_str(self) -> str:
        """
        Return the stop time of this Event as an ISO8601 string, such as 2024-11-07T15:30:45.12Z
        """
        return self._stop.isoformat().replace("+00:00", "Z")

    @property
    def name(self) -> str:
        """
        The name of the event, if this event is named, or "Unnamed Event" if not.
        """
        return self._name

    @staticmethod
    def from_dict(data_dict: dict):
        """
        Obtain an event from a dictionary.
        Dictionary must contain a key "start" containing an ISO8601 string or datetime.datetime object,
        a key "stop" containing an ISO8601 string or datetime.datetime object, and optionally a name.

        datetime.datetime objects and ISO8601 strings MUST contain timezone information.

        :param data_dict: valid dictionary containing data
        :return: an Event hydrated with the provided data
        """
        data = data_dict.copy()

        start_time = data["start"]
        del data["start"]

        end_time = data["stop"]
        del data["stop"]

        if "name" in data.keys():
            name = data["name"]
            del data["name"]

        else:
            name = None

        return Event(start_time, end_time, name, data)

    def to_dict(self) -> dict:
        """
        Compile this Event to its dictionary representation.

        :return: the representation of this Event as a dictionary
        """
        ret_dict = {
            "start": self.start_as_iso_str,
            "stop": self.stop_as_iso_str,
            "name": self.name
        }

        if self._attributes is not None:
            for key, value in self._attributes.items():
                ret_dict[key] = value

        return ret_dict



from data_tools.query.flux import FluxQuery
from data_tools.utils.times import ensure_utc
from data_tools.collections.time_series import TimeSeries
from pydantic import BaseModel, Field
from influxdb_client import InfluxDBClient
from dotenv import load_dotenv
from datetime import datetime, timedelta, timezone
import pandas as pd
import os

DEFAULT_INFLUXDB_URL = "http://influxdb.telemetry.ubcsolar.com"


class TimeSeriesTarget(BaseModel):
    """
    Encapsulates the data required to completely query a specific time series from InfluxDB.
    """
    name: str
    field: str
    measurement: str
    frequency: float = Field(gt=0)
    units: str
    car: str
    bucket: str
    description: str = Field(default_factory=str)


class DBClient:
    """
    This class encapsulates a connection to an InfluxDB database.
    """

    def __init__(self, influxdb_org=None, influxdb_token=None, url=None):
        """
        Create a connection to the InfluxDB database.

        :param influxdb_org: The name of the InfluxDB organization.
        :param influxdb_token: The API Token to the InfluxDB database.
        :param url: The URL to the InfluxDB database, default is "http://influxdb.telemetry.ubcsolar.com"
        """
        if influxdb_token is None or influxdb_org is None:
            load_dotenv()

            influxdb_token = os.getenv("INFLUX_TOKEN")
            influxdb_org = os.getenv("INFLUX_ORG")

        self._influx_org = influxdb_org
        url = DEFAULT_INFLUXDB_URL if url is None else url

        self._client = InfluxDBClient(url=url, org=influxdb_org, token=influxdb_token)

    def get_buckets(self) -> list[str]:
        """
        Obtain the available buckets.
        :return: list of available buckets
        """
        result = self._client.buckets_api().find_buckets()
        buckets = [bucket.name for bucket in result.buckets]
        return list(buckets)

    def get_measurements_in_bucket(self, bucket: str) -> list[str]:
        """
        Obtain a list of the possible measurements within a bucket.

        :param bucket: the bucket that will be queried. Must exist.
        :return: a list of available measurements in `bucket`.
        """
        query = f'''
                import "influxdata/influxdb/schema"
                schema.measurements(bucket: "{bucket}")
                '''

        result = self._client.query_api().query(org=self._influx_org, query=query)
        measurements = {record.get_value() for table in result for record in table.records}

        return list(measurements)

    def get_fields_in_measurement(self, bucket: str, measurement: str) -> list[str]:
        """
        Obtain the possible fields within a measurement and bucket.
        :param bucket: the bucket that will be queried. Must exist.
        :param measurement: the measurement that will be queried. Must exist.
        :return:
        """
        field_query = f'''from(bucket: "{bucket}")
                      |> range(start: -30d)
                      |> filter(fn: (r) => r["_measurement"] == "{measurement}")
                      |> group(columns: ["_field"])
                      |> distinct(column: "_field")'''

        field_result = self._client.query_api().query(org=self._influx_org, query=field_query)
        fields = {record.get_value() for table in field_result for record in table.records}

        return list(fields)

    def get_cars_in_measurement(self, bucket: str, measurement: str) -> list[str]:
        """
        Obtain the different cars within a measurement and bucket.
        :param bucket: the bucket that will be queried. Must exist.
        :param measurement: the measurement that will be queried. Must exist.
        :return:
        """
        field_query = f'''from(bucket: "{bucket}")
                      |> range(start: -30d)
                      |> distinct(column: "car")'''

        cars_result = self._client.query_api().query(org=self._influx_org, query=field_query)
        cars = {record.get_value() for table in cars_result for record in table.records}

        return list(cars)

    def query_dataframe(self, query: FluxQuery) -> pd.DataFrame:
        """
        Submit a Flux query, and return the result as a DataFrame.

        :param FluxQuery query: the query which will be submitted
        :return: the resulting data as a DataFrame
        """
        compiled_query = query.compile_query()
        compiled_query += ' |> pivot(rowKey:["_time"], columnKey: ["_field"], valueColumn: "_value") '

        return self._client.query_api().query_data_frame(compiled_query)

    def query_series(self, start: datetime, stop: datetime, field: str, bucket: str = "CAN_log",
                     car: str = "Brightside", measurement: str = None):
        """
        Query the database for a specific field, over a certain time range.
        The data will be returned as a DataFrame.

        :param start: the start time of the query as an ISO 8601-compliant string, such as "2024-06-30T23:00:00Z".
        :param stop: the end time of the query as an ISO 8601-compliant string, such as "2024-06-30T23:00:00Z".
        :param field: the field which is to be queried.
        :param str bucket: the bucket which will be queried
        :param car: the car which data is being queried for, default is "Brightside".
        :return: a TimeSeries of the resulting time-series data
        """
        # InfluxDB has an issue where PST timestamps were interpreted as UTC. So, we need to mutate
        # the timestamps to represent a time -7 hours to compensate for the UTC offset of +7.

        utc_start = ensure_utc(start) - timedelta(hours=7)
        utc_end = ensure_utc(stop) - timedelta(hours=7)

        # Make the query
        query = FluxQuery() \
            .from_bucket(bucket) \
            .range(start=utc_start.isoformat(), stop=utc_end.isoformat()) \
            .filter(field=field) \
            .car(car)

        if measurement:
            query = query.filter(measurement=measurement)
        query_df = self.query_dataframe(query)

        if isinstance(query_df, list):
            raise ValueError("Query returned multiple fields! Please refine your query.")

        if len(query_df) == 0:
            raise ValueError("Query is empty! Verify that the data is visible on InfluxDB for the queried bucket.")

        return query_df

    def query_time_series(self, start: datetime, stop: datetime, field: str, bucket: str = "CAN_log",
                          car: str = "Brightside", granularity: float = 0.1, units: str = "",
                          measurement: str = None) -> TimeSeries:
        """
        Query the database for a specific field, over a certain time range.
        The data will be processed into a TimeSeries, which has homogenous and evenly-spaced (temporally) elements.
        Data will be re-interpolated to have temporal granularity of ``granularity``.

        :param start: the start time of the query as an ISO 8601-compliant string, such as "2024-06-30T23:00:00Z".
        :param stop: the end time of the query as an ISO 8601-compliant string, such as "2024-06-30T23:00:00Z".
        :param field: the field which is to be queried.
        :param str bucket: the bucket which will be queried
        :param car: the car which data is being queried for, default is "Brightside".
        :param granularity: the temporal granularity of the resulting TimeSeries in seconds, default is 0.1s.
        :param units: the units of the returned data, optional.
        :return: a TimeSeries of the resulting time-series data
        """
        query_df = self.query_series(start, stop, field, bucket, car, measurement)

        return TimeSeries.from_query_dataframe(query_df, granularity, field, units)


if __name__ == "__main__":
    start_time = datetime(2024, 6, 16, 10, 0, 0, tzinfo=timezone.utc)
    end_time = datetime(2024, 8, 16, 23, 0, 0, tzinfo=timezone.utc)

    client = DBClient()

    pack_voltage: TimeSeries = client.query_time_series(start_time, end_time, "TotalPackVoltage", units="V",
                                                        measurement="BMS")

    pack_voltage.plot()



import pandas as pd
import pathlib
import datetime
import pytz


class FSGPDayLaps:

    def __init__(self, day: int):
        self.day: int = day

        self.filepath: pathlib.Path = (pathlib.Path(__file__).parent.parent
                         / "static_data"
                         / "fsgp_timing_2024"
                         / f"fsgp_timing_day_{day}.csv")
        self._df_full: pd.DataFrame = pd.read_csv(self.filepath)

        header: pd.Series = self._df_full.iloc[2]
        data: pd.DataFrame = self._df_full.iloc[4:]
        self.df = pd.DataFrame(data.values, columns=header)

        selected_cols: list = [
            'Lap #',
            'Start Time (Only if Diff than Prev Finish)',
            'Finish Time (HH:MM:SS)',
            'Lap Time (HH:MM:SS)',
            'Pit Time Before Lap (Min)',
            'Lap Time (Min)',
            'Avg Lap Speed (MPH)',
            'DRVR Name',
        ]

        dtypes: dict = {
            'Lap #': 'int',
            'Pit Time Before Lap (Min)': 'float',
            'Lap Time (Min)': 'float',
            'Avg Lap Speed (MPH)': 'float',
        }

        self.df: pd.DataFrame = self.df[selected_cols].dropna(axis=0, how='all').astype(dtypes)
        self.df.set_index('Lap #', inplace=True)

    @staticmethod
    def _pad_timestamp(timestamp: str) -> str:
        """
        Pad a string timestamp to match HH:MM:SS format

        :param timestamp: timestamp with H:MM:SS or HH:MM:SS format
        :return: timestamp in HH:MM:SS format
        """
        assert timestamp.count(":") == 2, "timestamp is not in HH:MM:SS format - should have 2 colons"
        if len(timestamp) == 8:
            return timestamp
        elif len(timestamp) == 7:
            return "0" + timestamp
        else:
            raise ValueError('timestamp did not have H:MM:SS format or HH:MM:SS format, should have len 7 or 8')

    def get_lap_count(self) -> int:
        return self.df.index.max()

    def _get_utc_str(self, time_str: str) -> str:

        date_time_str = f"{'2024-07-'}{15 + self.day} {time_str}"
        naive_datetime = datetime.datetime.strptime(date_time_str, "%Y-%m-%d %H:%M:%S")

        central_tz = pytz.timezone('America/Chicago')
        local_datetime = central_tz.localize(naive_datetime)
        utc_datetime = local_datetime.astimezone(pytz.utc)

        # Format the UTC time to ISO 8601 string format
        return utc_datetime.strftime("%Y-%m-%dT%H:%M:%S") + "Z"

    def get_start_utc(self, lap: int) -> datetime:
        """
        Get lap start time as a UTC datetime object

        :param lap: Lap number from the race day
        :return: Start time as a UTC datetime object
        """
        timestamp_str = self.get_start_utc_string(lap)
        # remove last "Z" char
        return (datetime.datetime.strptime(timestamp_str[:-1], "%Y-%m-%dT%H:%M:%S")
                .replace(tzinfo=datetime.timezone.utc))

    def get_finish_utc(self, lap: int) -> datetime:
        """
        Get lap finish time as a UTC datetime object

        :param lap: Lap number from the race day
        :return: Finish time as a UTC datetime object
        """
        timestamp_str = self.get_finish_utc_string(lap)
        # remove last "Z" char
        return (datetime.datetime.strptime(timestamp_str[:-1], "%Y-%m-%dT%H:%M:%S")
                .replace(tzinfo=datetime.timezone.utc))

    def get_start_utc_string(self, lap: int) -> str:
        """
        Get lap start time as a UTC timestamp

        :param lap: Lap number from the race day
        :return: Start time in %Y-%m-%dT%H:%M:%SZ format e.g. 2024-07-16T20:46:32Z
        """
        start_time = self.df.loc[lap, 'Start Time (Only if Diff than Prev Finish)']
        assert lap > 0, "Lap number must be greater than zero; first lap is lap 1"
        if isinstance(start_time, str) and len(start_time) > 0:
            time_str = start_time
        else:
            time_str = self.df.loc[lap - 1, 'Finish Time (HH:MM:SS)']
        return self._get_utc_str(time_str)

    def get_finish_utc_string(self, lap: int) -> str:
        """
        Get lap finish time as a UTC timestamp

        :param lap: Lap number from the race day
        :return: Finish time in %Y-%m-%dT%H:%M:%SZ format e.g. 2024-07-16T20:46:32Z
        """
        assert lap > 0, "Lap number must be greater than zero; first lap is lap 1"
        time_str = self.df.loc[lap, 'Finish Time (HH:MM:SS)']
        return self._get_utc_str(time_str)

    def get_time(self, lap: int) -> str:
        """
        Get lap time as an HH:MM:SS timestamp

        :param lap: Lap number from the race day
        :return: Time in HH:MM:SS format e.g. 00:06:59
        """
        assert lap > 0, "Lap number must be greater than zero; first lap is lap 1"
        return self._pad_timestamp(self.df.loc[lap, 'Lap Time (HH:MM:SS)'])

    def get_pit_time(self, lap: int) -> float:
        """
        Get pit time before lap in minutes as a decimal, or 0. if Brightside didn't pit

        :param lap: Lap number from the race day
        :return: Pit time before lap in minutes as a decimal, e.g. 3.583
        """
        assert lap > 0, "Lap number must be greater than zero; first lap is lap 1"
        return self.df.loc[lap, 'Pit Time Before Lap (Min)']

    def get_time_minutes(self, lap: int) -> float:
        """
        Get lap time in minutes as a decimal

        :param lap: Lap number from the race day
        :return: Time in minutes as a decimal, e.g. 6.324
        """
        assert lap > 0, "Lap number must be greater than zero; first lap is lap 1"
        return self.df.loc[lap, 'Lap Time (Min)']

    def get_lap_mph(self, lap: int) -> float:
        """
        Get the average lap speed in mph

        :param lap: Lap number from the race day
        :return: Speed in mph, e.g. 26.745
        """
        assert lap > 0, "Lap number must be greater than zero; first lap is lap 1"
        return self.df.loc[lap, 'Avg Lap Speed (MPH)']

    def get_lap_driver(self, lap: int) -> str:
        """
        Get the name of a driver for a lap

        :param lap: Lap number from the race day
        :return: Name of driver, e.g. 'Diego'
        """
        assert lap > 0, "Lap number must be greater than zero; first lap is lap 1"
        return self.df.loc[lap, 'DRVR Name']



from datetime import datetime, timezone


def ensure_utc(dt: datetime) -> datetime:
    """
    Ensure that a datetime, ``dt`` is localized to UTC.

    :param dt: the datetime that will be validated
    :raises ValueError: if ``dt`` is not localized to ANY timezone.
    :return:
    """
    # Check if ``dt`` is naive (not localized to a timezone), in that case we cannot safely proceed.
    if dt.tzinfo is None:
        raise ValueError("Datetime object must be timezone-aware.")

    # Otherwise, we can re-localize the ``dt`` to UTC if it isn't already
    if dt.tzinfo != timezone.utc:
        dt = dt.astimezone(timezone.utc)

    return dt


def parse_iso_datetime(iso_string: str) -> datetime:
    """
    Convert an ISO 8601 string to a timezone-aware datetime object.

    :param str iso_string: ISO 8601 formatted datetime string (e.g. "2024-01-01T15:30:00Z" or "2024-01-01T15:30:00+00:00")

    :returns: timezone-aware datetime object
    :rtype: datetime
    :raises ValueError: If the ISO string is invalid or timezone name is not recognized
    """
    try:
        # Handle both 'Z' and '+00:00' UTC formats
        if iso_string.endswith('Z'):
            dt = datetime.fromisoformat(iso_string.replace('Z', '+00:00'))
        else:
            dt = datetime.fromisoformat(iso_string)

        dt_utc = ensure_utc(dt)

        return dt_utc

    except Exception as e:
        raise ValueError(f"Failed to parse ISO datetime string: {str(e)}") from e


def iso_string_from_datetime(dt: datetime) -> str:
    """

    Return the stop time of this Event as an ISO8601 string, such as 2024-11-07T15:30:45.12Z

    :param dt: datetime object to be converted
    :return: the datetime as an ISO8601 string
    """
    return dt.isoformat().replace("+00:00", "Z")



"""
=================================
Schema (:mod:`data_tools.schema`)
=================================

Schema is a module which composes a set of classes and objects
to abstractly construct a filesystem-like API

File API
========

.. autosummary::
   :toctree: generated/

   File             -- Atomic unit of data
   CanonicalPath    -- Represent a location in an abstract filesystem
   FileType         -- Discretize different file types


Data Handling
=============

.. autosummary::
   :toctree: generated/

   Result           -- Algebraic datatype that can either represent a successful result of some operation, or a failure.
   UnwrappedError   -- Error arising from unwrapping a ``Result`` which contains a failure


DataSource API
==============

.. autosummary::
   :toctree: generated/

   DataSource      -- Abstract interface for storing and acquiring data
   FileLoader      -- A useful callable mapping to a stored File in an abstract filesystem
   Event           -- A named time range

"""


from ._event import (
    Event
)

from ._result import (
    UnwrappedError,
    Result
)

from ._file import (
    FileType,
    File,
    CanonicalPath
)

from ._file_loader import (
    FileLoader
)

from ._data_source import (
    DataSource
)

__all__ = [
    "DataSource",
    "File",
    "FileType",
    "FileLoader",
    "CanonicalPath",
    "Result",
    "UnwrappedError",
    "Event"
]



"""
ORM-like query-building tools for Flux, InfluxDB's custom querying language.

See InfluxDB's documentation for more details: https://docs.influxdata.com/influxdb/v2/query-data/flux/
"""
from typing import Callable, List
from dateutil import parser


# Function signature required for a parameterized Flux statement formatter
FluxFormatter = Callable[[List[str]], str]


class FluxStatement:
    """
    Represent and construct a single Flux statement. such as
    ``"filter(fn:(r) => r._field == "TotalPackVoltage")``.

    A FluxStatement is the atomic unit that can be composed into a FluxQuery, which is capable of querying our
    InfluxDB database.
    """
    def __init__(self, statement: str, priority: int):
        assert isinstance(statement, str), f"Statement must be of type `str`, not {type(statement)}!"
        assert (isinstance(priority, int) and priority >= 0), "Priority must be a non-negative integer!"

        self._statement: str = statement
        self._priority: int = priority

    @property
    def priority(self) -> int:
        """
        The priority of this statement, which determines its position in a Flux query. Lower number indicates
        higher priority.
        """
        return self._priority

    @property
    def statement(self) -> str:
        """
        The canonical string representation of this Flux query.
        """
        return self._statement


class ParameterizedFluxStatement:
    """
    Represent and encapsulate a parameterized Flux statement.

    An instance of this class can be instantiated statically as a part
    of a module to provide built-in support for standard Flux statements.
    """
    def __init__(self, formatter: FluxFormatter, priority: int):
        """
        Instantiate a ParameterizedFluxStatement.

        :param FluxFormatter formatter: must be a ``Callable`` that takes in a ``List[str]``, the parameters to the Flux query, and returns a ``str``, which is the Flux query injected with parameters.
        :param int priority: must be a non-negative integer that indicates the priority of the Flux statement encapsulated by this class.
        """
        assert isinstance(formatter, Callable), "formatter must have signature `Callable[[List[str]], str]`!"
        assert (isinstance(priority, int) and priority >= 0), "Priority must be a non-negative integer!"

        self._formatter: FluxFormatter = formatter
        self._priority: int = priority

    def get(self, parameters: List[str]) -> FluxStatement:
        """
        Inject the ``parameters`` into this ``ParameterizedFluxStatement`` to produce a concrete ``FluxStatement``.

        :param List[str] parameters: the parameters to be injected, must match the number of paramaters expected by the formatter used to construct this ``ParameterizedFluxStatement``.
        :return: the concrete ``FluxStatement``
        """
        return FluxStatement(self._formatter(parameters), self.priority)

    @property
    def priority(self) -> int:
        """
        Obtain the priority that concrete ``FluxStatement``s produced by this ``ParameterizedFluxStatement`` will inherit.
        """
        return self._priority


class FluxQuery:
    """
    Represent a Flux query: a collection of ``FluxStatement`` chained together to form a complete query
    to our InfluxDB database.

    Supports method chaining for fluent construction of queries.
    """
    _from_bucket_formatter: FluxFormatter = lambda args: f'from(bucket: "{args[0]}") '
    _from_bucket_statement = ParameterizedFluxStatement(formatter=_from_bucket_formatter, priority=0)

    _unbound_range_formatter: FluxFormatter = lambda args: f'range(start: {args[0]}) '
    _unbound_range_statement = ParameterizedFluxStatement(formatter=_unbound_range_formatter, priority=1)

    _bound_range_formatter: FluxFormatter = lambda args: f'range(start: {args[0]}, stop: {args[1]}) '
    _bound_range_statement = ParameterizedFluxStatement(formatter=_bound_range_formatter, priority=1)

    _filter_measurement_formatter: FluxFormatter = lambda args: f'filter(fn:(r) => r._measurement == "{args[0]}") '
    _filter_measurement_statement = ParameterizedFluxStatement(formatter=_filter_measurement_formatter, priority=2)

    _filter_field_formatter: FluxFormatter = lambda args: f'filter(fn:(r) => r._field == "{args[0]}") '
    _filter_field_statement = ParameterizedFluxStatement(formatter=_filter_field_formatter, priority=3)

    _car_formatter: FluxFormatter = lambda args: f'filter(fn:(r) => r.car == "{args[0]}") '
    _car_statement = ParameterizedFluxStatement(formatter=_car_formatter, priority=3)

    def __init__(self):
        """
        Instantiate an empty Flux query.
        :return: this FluxQuery with the new statement inserted
        """
        self._statements: List[FluxStatement] = []

    def from_bucket(self, bucket: str):
        """
        Specify the bucket that will be queried from.

        :param str bucket: bucket name
        :return: this FluxQuery with the new statement inserted
        """
        assert isinstance(bucket, str), f"Bucket must be a `str`, not {type(bucket)}!"

        new_statement = FluxQuery._from_bucket_statement.get([bucket])
        self._statements.append(new_statement)

        return self

    def range(self, start: str, stop: str = None):
        """
        Specify the time range of the time-series data to query.

        :param start: start time of the time range as an ISO 8601 compliant string.
        :param stop: stop time of the time range as an ISO 8601 compliant string, optional.
        :return: this FluxQuery with the new statement inserted
        """
        # Verify that `start` and `stop` are valid ISO 8601 dates.
        try:
            parser.parse(start)
            if stop is not None:
                parser.parse(stop)
        except parser.ParserError:
            raise ValueError("Invalid dates provided to range! Must be of ISO 8601 format.")

        if stop is not None:
            new_statement = FluxQuery._bound_range_statement.get([start, stop])
            self._statements.append(new_statement)

        else:
            new_statement = FluxQuery._unbound_range_statement.get([start])
            self._statements.append(new_statement)

        return self

    def filter(self, measurement: str = None, field: str = None):
        """
        Apply a filter to the query. Any or both of ``measurement`` and ``field`` may be specified and both filters will
        be applied. No effect if neither is.

        :param str measurement: measurement to filter for, such as "BMS"
        :param str field: field to filter for, such as "PackCurrent".
        :return: this FluxQuery with the new statement inserted
        """
        if measurement is not None:
            new_statement = FluxQuery._filter_measurement_statement.get([measurement])
            self._statements.append(new_statement)

        if field is not None:
            new_statement = FluxQuery._filter_field_statement.get([field])
            self._statements.append(new_statement)

        return self

    def car(self, car: str):
        """
        Filter for data from a specific car.

        :param car: car name, such as "Brightside".
        :return: this FluxQuery with the new statement inserted
        """
        assert isinstance(car, str), f"Car must be a `str`, not {type(car)}!"

        new_statement = FluxQuery._car_statement.get([car])
        self._statements.append(new_statement)

        return self

    def inject_raw(self, statement: FluxStatement):
        """
        Inject a FluxStatement into this query. This may be necessary if the
        Flux statement is not available with the existing built-in Flux statements.

        :param statement:
        :return: this FluxQuery with the new statement inserted
        """
        self._statements.append(statement)

        return self

    def compile_query(self) -> str:
        """
        Compile this object into a Flux query, as a string.

        :return: the canonical string representation of this Flux query, with statements ordered according to priority.
        """
        sorted_statements = sorted(self._statements, key=lambda statement: statement.priority)

        compiled_statement = sorted_statements.pop(0).statement
        for flux_statement in sorted_statements:
            compiled_statement += f"|> {flux_statement.statement}"

        return compiled_statement



"""
===========================================
Collections (:mod:`data_tools.lap_tools`)
===========================================

Race Data Tools
===========
.. autosummary::
   :toctree: generated/

   FSGPDayLaps    -- Data parser and container for FSGP 2024 lap data

.. autosummary::
   :toctree: generated/

   collect_lap_data    -- Collects data over each lap using FSGP 2024 timestamps

"""


from .fsgp_2024_laps import FSGPDayLaps
from .lap_query import collect_lap_data


__all__ = [
    "FSGPDayLaps",
    "collect_lap_data"
]



from enum import StrEnum
from typing import List, Any
from functools import reduce
from pydantic import BaseModel, Field, ConfigDict
import pathlib


class FileType(StrEnum):
    """
    Discretize the valid types of data that a `File` is supported to contain.
    All `DataSource` implementations can be expected to satisfy each supported file type for all applicable methods.
    """
    TimeSeries = "TimeSeries"
    NDArray = "NDArray"
    Scalar = "Scalar"
    Empty = "Empty"
    Any = "Any"


class CanonicalPath:
    def __init__(self, origin: str, source: str, event: str, name: str):
        """
        Construct a canonical path representing a path to a file in any abstract data source.

        :param str origin: Identifies the origin (code) of this data, usually the data pipeline version.
        :param str source: The producer of the data pointed to by this canonical path, usually a pipeline stage
        :param str event: The event that this data belongs to
        :param str name: The name of this data
        """
        self._origin: str = origin
        self._source: str = source
        self._event: str = event
        self._name: str = name

    def to_string(self) -> str:
        """
        Obtain the string representation of this canonical path
        """
        return "/".join([self._origin, self._event, self._source, self._name])

    def to_path(self) -> pathlib.Path:
        return reduce(lambda x, y: x / pathlib.Path(y), self.unwrap())

    @property
    def origin(self) -> str:
        return self._origin

    @property
    def source(self) -> str:
        return self._source

    @property
    def event(self) -> str:
        return self._event

    @property
    def name(self) -> str:
        return self._name

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string()

    def unwrap(self) -> List[str]:
        """
        Decompose this `CanonicalPath` into its constituent elements. Equivalent to `os.path.split`.
        """
        return self.to_string().split("/")

    @staticmethod
    def unwrap_canonical_path(canonical_path: str) -> List[str]:
        """
        Unwrap a canonical path into its elements.

        For example, `"pipeline_2024_11_01/ingest/TotalPackVoltage"` would be
        unwrapped to `["pipeline_2024_11_01", "ingest", "TotalPackVoltage"]`.

        The first element should always be a reference to the origin (code) that produced this data.
        The second element should always refer to the stage (processing step) that produced this data.
        The last element should always be the name of this data.

        :param canonical_path: The path to be decomposed
        :return: A List[str] of path elements
        """
        return canonical_path.split("/")


class File(BaseModel):
    """
    An atomic unit of data, described by data, a file type describing the data stored, and a canonical
    path denoting its location in some filesystem-like storage.
    """
    data: Any                                       # The data contained by this file
    file_type: FileType                             # The type of data contained by this file
    canonical_path: CanonicalPath                   # The path to this file within the data source where it is stored
    metadata: dict = Field(default_factory=dict)    # Any additional metadata that this file should hold
    description: str = Field(default_factory=str)   # A description of what this file is

    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)



from abc import ABC, abstractmethod
from data_tools.schema import Result, File, FileLoader, CanonicalPath


class DataSource(ABC):
    """
    Abstract base class for data sources.
    """
    def __init__(self, *args, **kwargs) -> None:
        pass

    @abstractmethod
    def store(self, file: File) -> FileLoader:
        raise NotImplementedError

    @abstractmethod
    def get(self, canonical_path: CanonicalPath, **kwargs) -> Result:
        raise NotImplementedError



import numpy as np
import datetime
from dateutil import parser
import matplotlib.pyplot as plt
import math
from warnings import warn
import pandas as pd
import re


class TimeSeries(np.ndarray):
    """
    This class encapsulates time-series data with units, a temporal x–axis, and metadata.

    Data is homogenous and evenly-spaced, such that temporal granularity between subsequent elements is constant.

    TimeSeries can be indexed with a ``float`` or slice with ``float`` components in order to index by relative time.
    For example, for some ``timeSeries``, ``timeSeries[10.43]`` is equivalent to
    ``timeSeries[timeSeries.index_of(10.43)]``.
    """

    # __new__ and __array_finalize__ are mandatory to ensure that
    # `TimeSeries` properly acts like a ndarray when necessary.
    def __new__(cls, input_array, meta):
        obj = np.asarray(input_array).view(cls)
        obj._meta = meta
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self._start = getattr(obj, '_start', None)
        self._stop = getattr(obj, '_stop', None)
        self._units = getattr(obj, '_units', None)
        self._length = getattr(obj, '_length', None)
        self._period = getattr(obj, '_period', None)
        self._meta = getattr(obj, '_meta', None)

    def __init__(self, input_array, meta: dict):
        assert isinstance(meta["start"], type(meta["stop"])), "Start and stop times are not of same type!"
        if isinstance(meta["start"], datetime.datetime) and isinstance(meta["stop"], datetime.datetime):
            self._start: datetime.datetime = meta["start"]
            self._stop: datetime.datetime = meta["stop"]
        else:
            print(meta["start"])
            self._start: datetime.datetime = parser.parse(meta["start"])
            self._stop: datetime.datetime = parser.parse(meta["stop"])
        del meta["start"]
        del meta["stop"]

        self._units: str = meta["units"]
        del meta["units"]

        if "period" in meta.keys():
            self._period: float = meta["period"]
            del meta["period"]

        else:
            warn("Please set `period` and not granularity. It will be removed in the future.", DeprecationWarning)
            self._period: float = meta["granularity"]
            del meta["granularity"]

        self._length: float = meta["length"]
        del meta["length"]

        self._meta: dict = meta

    @property
    def x_axis(self) -> np.ndarray:
        """
        This wave's x–axis in relative seconds, such that the first element is ``t=0``.
        """
        relative_x_axis = np.linspace(0.0, self._length, len(self))

        return relative_x_axis

    @property
    def unix_x_axis(self) -> np.ndarray:
        """
        This wave's x–axis as UTC UNIX timestamps.
        """
        return self.x_axis + self.start.timestamp()

    @property
    def datetime_x_axis(self) -> np.ndarray:
        """
        This wave's x–axis as ``datetime``s, in UTC.
        """

        return pd.to_datetime(self.unix_x_axis, unit='s')

    @property
    def length(self) -> float:
        """
        Total time between the first and last element of this wave's data
        """

        return self._length

    @property
    def granularity(self) -> float:
        """
        Temporal granularity (delta t) between each element of this wave's data in seconds
        """
        warn("Please use TimeSeries.period instead of TimeSeries.granularity", DeprecationWarning, stacklevel=2)
        return self._period

    @property
    def period(self) -> float:
        """
        Get the period, in seconds, between data points of this TimeSeries.

        :return: period, in seconds.
        """
        return self._period

    @property
    def units(self) -> str:
        """
        The units of this wave's data
        """
        return self._units

    @units.setter
    def units(self, value: str):
        assert isinstance(value, str), f"Units should be a string, not {type(value)}!"
        self._units = value

    @property
    def start(self) -> datetime.datetime:
        """
        UTC datetime of the first data element
        """
        return self._start

    @property
    def stop(self) -> datetime.datetime:
        """
        UTC datetime of the last data element
        """
        return self._stop

    def __getitem__(self, item):
        data_to_slice = self

        if isinstance(item, float):
            index = self.index_of(item)
            return super(TimeSeries, data_to_slice).__getitem__(index)

        elif isinstance(item, slice):
            if isinstance(item.start, float):
                start_index = self.index_of(item.start)
            else:
                start_index = item.start

            if isinstance(item.stop, float):
                stop_index = self.index_of(item.stop)
            else:
                stop_index = item.stop

            item = slice(start_index, stop_index, item.step)

            unix_x_axis = data_to_slice.unix_x_axis
            new_start_timestamp: float = unix_x_axis[start_index]
            new_stop_timestamp: float = unix_x_axis[stop_index - 1]

            new_start: datetime.datetime = datetime.datetime.fromtimestamp(new_start_timestamp)
            new_stop: datetime.datetime = datetime.datetime.fromtimestamp(new_stop_timestamp)
            new_length: float = new_stop_timestamp - new_start_timestamp

            new_time_series = TimeSeries(self, {
                "start": new_start,
                "stop": new_stop,
                "car": data_to_slice.meta["car"],
                "measurement": data_to_slice.meta["measurement"],
                "field": data_to_slice.meta["field"],
                "period": data_to_slice.period,
                "length": new_length,
                "units": data_to_slice.units,
            })

            data_to_slice = new_time_series

        return super(TimeSeries, data_to_slice).__getitem__(item)

    @property
    def meta(self) -> dict:
        """
        Metadata such as the field, measurement, and car.
        """
        return self._meta

    @meta.setter
    def meta(self, new_meta: dict):
        assert isinstance(new_meta, dict), f"New metadata should be a dictionary, not {type(new_meta)}!"
        self._meta = new_meta

    def plot(self, show=True) -> None:
        """
        Make a simple plot this data.
        :param bool show: Show plots (disable if you want to stack multiple plots, for example).
        """

        fig, ax = plt.subplots()

        ax.set_title(f"{self.meta['measurement']}: {self.meta['field']}")
        ax.set_ylabel(self.units if self.units != "" else "Arbitrary Units")
        ax.set_xlabel("Time (s)")
        ax.plot(self.datetime_x_axis, self, label=self.meta['field'])

        if show:
            plt.show()

    def index_of(self, time: float) -> int:
        """
        Return the index of the data element that represents the time closest to ``time``.

        :param float time: time (in seconds) that will be evaluated against
        :raises: ValueError if ``time`` falls outside of x–axis
        """
        if not (0.0 <= time <= self.length):
            raise ValueError(f"Relative time {time} falls outside of x–axis!")

        return np.argmin(np.abs(self.x_axis - time))

    def relative_time(self, unix_time: float) -> float:
        """
        Return the relative time of the UNIX timestamp ``time``.
        :param float unix_time: UNIX timestamp that will be converted
        """
        if not (self.start.timestamp() <= unix_time <= self.stop.timestamp()):
            raise ValueError(f"UNIX time {unix_time} falls outside of x–axis, which is {self.start.timestamp()}–{self.stop.timestamp()}!")

        return unix_time - self.start.timestamp()

    def promote(self, array: np.ndarray):
        """
        Promote a plain ndarray, ``array``, to a TimeSeries with the same metadata
        as this TimeSeries.

        This method is particularly useful for interfacing
        with libraries such as SciPy and NumPy, which will return an ndarray even when
        given a TimeSeries.

        :param array: plain ndarray to be promoted
        :return: new, promoted TimeSeries with the same metadata as this TimeSeries
        """
        meta: dict = {
            "start": self.start,
            "stop": self.stop,
            "car": self.meta["car"],
            "measurement": self.meta["measurement"],
            "field": self.meta["field"],
            "period": self.period,
            "length": self.length,
            "units": self.units,
        }

        return TimeSeries(array, meta)

    @staticmethod
    def align(*args) -> list:
        start_time = max(arg.start.timestamp() for arg in args)
        end_time = min(arg.stop.timestamp() for arg in args)
        period = min(arg.period for arg in args)

        new_length = end_time - start_time
        num_points = math.ceil(new_length / period) + 1
        new_x_axis = np.linspace(start_time, end_time, num_points)

        new_args = []
        for array in args:
            start_index = array.index_of(array.relative_time(start_time))
            stop_index = array.index_of(array.relative_time(end_time))

            new_array = array[start_index:stop_index + 1]
            new_array._start = datetime.datetime.fromtimestamp(start_time)
            new_array._stop = datetime.datetime.fromtimestamp(end_time)
            new_array._period = period
            new_array._length = new_length

            interpolated_values = np.interp(new_x_axis, new_array.unix_x_axis, new_array)
            new_array_interpolated = new_array.promote(interpolated_values)

            new_args.append(new_array_interpolated)

        return new_args

    @staticmethod
    def from_query_dataframe(query_df: pd.DataFrame, granularity: float, field: str, units: str):
        # Transform the DataFrame into a nicer format where we have our time-series data indexed by time
        query_df['_time'] = pd.to_datetime(query_df['_time'])
        query_df.set_index('_time', inplace=True)

        # Get the x-axis in relative seconds (first element is t=0)
        x_axis = query_df.index.map(lambda x: x.timestamp()).to_numpy()
        x_axis -= x_axis[0]  # Subtract off first time, so the x_axis starts at 0 with units of seconds

        # Reshape the x-axis to have the right number of elements for our needed granularity
        temporal_length: float = x_axis[-1]  # Total time of the query in seconds
        desired_num_elements: int = math.ceil(temporal_length / granularity)
        desired_x_axis = np.linspace(0, temporal_length, desired_num_elements, endpoint=True)

        # Re-interpolate our data on desired x-axis
        wave = query_df[[field]].to_numpy().reshape(-1)
        wave_interpolated = np.interp(desired_x_axis, x_axis, wave)

        actual_granularity = np.mean(np.diff(desired_x_axis))

        # Compile metadata
        meta: dict = {
            "start": query_df.index.to_numpy()[0].to_pydatetime(),
            "stop": query_df.index.to_numpy()[-1].to_pydatetime(),
            "car": query_df["car"].to_numpy()[0],
            "measurement": query_df["_measurement"].to_numpy()[0],
            "field": field,
            "period": actual_granularity,
            "length": temporal_length,
            "units": units,
        }

        new_wave = TimeSeries(wave_interpolated, meta)

        return new_wave

    @staticmethod
    def from_csv(path, granularity, field):
        df = pd.read_csv(path)

        fields = df['_field'].unique()
        assert len(fields) == 1, "more than one field type in table"
        tuple_arr = np.stack((pd.to_datetime(df['_time']), df['_value']), axis=1)
        result = np.fromiter(map(lambda x: re.sub("[^0-9|.]", "", x), tuple_arr[:, 1]), dtype=float)

        return result



from enum import StrEnum
from typing import TypeVar


T = TypeVar("T")


class UnwrappedError(Exception):
    pass


class Result[T]:
    """
    Algebraic datatype that can either represent a successful result of some operation,
    or a failure.

    A `Result` can be unwrapped to retrieve the result or raise the exception raised wrapped in a `UnwrappedError`.
    """
    class ResultType(StrEnum):
        Ok = "Ok"
        Err = "Err"

    def __init__(self, result: T, result_type: ResultType):
        """
        Do not call this directly.
        Call `Result.Ok()` or `Result.Err()`.

        :param T result: Data or exception to be wrapped.
        :param ResultType result_type: Whether the result is data or an exception.
        :raises TypeError: If trying to wrap data as an `Exception` or an `Exception` as data
        """
        if result_type == Result.ResultType.Ok and isinstance(result, Exception):
            raise TypeError("Cannot wrap an Exception with Result.Ok()! Wrap it with Result.Err() instead.")

        if not isinstance(result, Exception) and result_type == Result.ResultType.Err:
            raise TypeError("Cannot wrap a non-Exception with Result.Err()! Wrap it with Result.Ok() instead.")

        self._result = result
        self._result_type = result_type

    @staticmethod
    def Ok(result: T):
        """
        Wrap a successful result in a `Result`.

        :param result: The successful result to be wrapped
        :return: A `Result` instance wrapping `result`.
        :raises TypeError: If `result` is an `Exception`.
        """
        return Result(result, Result.ResultType.Ok)

    @staticmethod
    def Err(error: Exception):
        """
        Wrap an error/failure/exception in a `Result`.

        :param error: The error to be wrapped
        :return: A `Result` instance wrapping `error`.
        :raises TypeError: If `error` is not an `Exception`.
        """
        return Result(error, Result.ResultType.Err)

    def unwrap(self) -> T | UnwrappedError:
        """
        Unwrap this `Result` to reveal a successful result or an error.

        :raises UnwrappedError: If an error is unwrapped
        :return: The result, if it was successful
        """
        if self._result_type == self.ResultType.Ok:
            return self._result

        # We must have self._result_type == Result.ResultType.Err
        raise UnwrappedError from self._result

    def __bool__(self):
        return True if self._result_type == self.ResultType.Ok else False



"""
===========================================
Querying Tools (:mod:`data_tools.query`)
===========================================

Flux Tools
==========

.. autosummary::
   :toctree: generated/

   FluxStatement      -- Atomic component of FluxQuery
   FluxQuery          -- Query composed of FluxStatements

Database Tools
==============

.. autosummary::
   :toctree: generated/

   DBClient           -- Powerful and simple InfluxDB client
   SunbeamClient      -- A simple client for accessing Sunbeam's API, UBC Solar's custom data pipeline
   PostgresClient     -- Powerful and simple PostgreSQL client

"""


from .flux import FluxQuery, FluxStatement
from .influxdb_query import DBClient, TimeSeriesTarget
from .postgresql_query import PostgresClient
from .data_schema import get_sensor_id, get_data_units, CANLog, init_schema
from ._sunbeam import SunbeamClient, SunbeamCache

__all__ = [
    "FluxQuery",
    "FluxStatement",
    "DBClient",
    "PostgresClient",
    "TimeSeriesTarget",
    "get_sensor_id",
    "get_data_units",
    "init_schema",
    "SunbeamClient",
    "SunbeamCache",
    "CANLog"
]



from data_tools.query.data_schema import init_schema, CANLog, get_sensor_id, get_data_units
from data_tools.collections.time_series import TimeSeries
from sqlalchemy.orm import sessionmaker, Session
from data_tools.utils.times import ensure_utc
from sqlalchemy import create_engine, Engine
from datetime import datetime, timezone
from typing import List, Type, Union
import matplotlib.pyplot as plt
from dotenv import load_dotenv
import numpy as np
import logging
import os

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)  # Set the logger to the lowest level (DEBUG)

# Create a handler for terminal output
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)  # Show INFO and ERROR on the terminal
logger.addHandler(console_handler)

load_dotenv()

_POSTGRESQL_USERNAME = os.getenv("POSTGRESQL_USERNAME")
_POSTGRESQL_PASSWORD = os.getenv("POSTGRESQL_PASSWORD")
_POSTGRESQL_DATABASE = os.getenv("POSTGRESQL_DATABASE")
_POSTGRESQL_ADDRESS = os.getenv("POSTGRESQL_ADDRESS")

if _POSTGRESQL_USERNAME is None:
    logger.error("Could not find a POSTGRESQL_USERNAME in .env!")

if _POSTGRESQL_PASSWORD is None:
    logger.error("Could not find a POSTGRESQL_PASSWORD in .env!")

if _POSTGRESQL_DATABASE is None:
    logger.error("Could not find a POSTGRESQL_DATABASE in .env!")

if _POSTGRESQL_ADDRESS is None:
    logger.error("Could not find a POSTGRESQL_ADDRESS in .env!")


def _get_db_url(db_name: str, ip_address: str, username: str, password: str) -> str:
    """
    Get the URL to a Postgres database.

    :param str db_name: the name of the database being connected to
    :param str ip_address: the IP Address of the machine running the PostgreSQL instance being connected to
    :param str username: the username of the user that is connecting
    :param str password: the password of the user that is connecting
    :return: the URL formatted as a string
    """
    assert isinstance(db_name, str), f"db_name must be a string, not {type(db_name)}!"
    assert isinstance(username, str), f"username must be a string, not {type(username)}!"
    assert isinstance(ip_address, str), f"ip_address must be a string, not {type(ip_address)}!"
    assert isinstance(password, str), f"password must be a string, not {type(password)}!"

    return f"postgresql://{username}:{password}@{ip_address}:5432/{db_name}"


class PostgresClient:
    """

    Connect to a PostgresSQL database and concisely make queries for time-series data.

    """
    def __init__(self, db_name: str = None, ip_address: str = None, username: str = None, password: str = None, timeout: float = 10):
        if username is None:
            username = _POSTGRESQL_USERNAME
        if db_name is None:
            db_name = _POSTGRESQL_DATABASE
        if ip_address is None:
            ip_address = _POSTGRESQL_ADDRESS
        if password is None:
            password = _POSTGRESQL_PASSWORD

        url = _get_db_url(db_name, ip_address, username, password)
        self._engine: Engine = create_engine(url, connect_args={'connect_timeout': timeout})

        self._session_builder = sessionmaker(bind=self._engine)
        self._session: Session = self._session_builder()

    def query(self, field: str, start_time: datetime, end_time: datetime, granularity: float = 1.0) -> TimeSeries:
        """
        Query the database for time-series data matching ``field``, between ``start_time`` and ``end_time``.

        Undefined behaviour when timestamps provided are not in UTC!

        PLEASE NOTE: If your data does not have a frequency of 1.0Hz, then you will be
        decimating or up-sampling your data. Set ``granularity`` to control the temporal
        granularity (time between measurements) of your data!

        :param str field: name of the field that will be queried
        :param datetime.datetime start_time: the UTC datetime of the beginning of the data to be queried
        :param datetime.datetime end_time: the UTC datetime of the end of the data to be queried
        :param float granularity: the desired temporal granularity (time between measurements) of the returned data
        :return: successfully queried data formatted as a TimeSeries
        :raises IndexError: if no data could be queried
        :raises RuntimeError: if the query failed for any reason
        :raises ValueError: if timestamps cannot be localized to UTC
        """
        # Validate that the times provided are localized to UTC
        utc_start_time = ensure_utc(start_time)
        utc_end_time = ensure_utc(end_time)

        unix_start_time = utc_start_time.timestamp()
        unix_end_time = utc_end_time.timestamp()

        try:
            queried_data: List[Type[CANLog]] = self._session.query(CANLog)\
                .filter(CANLog.sensor_type == get_sensor_id(field),
                        CANLog.timestamp >= unix_start_time,
                        CANLog.timestamp <= unix_end_time)\
                .order_by(CANLog.timestamp).all()

        # Catch any error from querying and wrap it in a RuntimeError
        except Exception as e:
            raise RuntimeError("Error whilst querying data!") from e

        if len(queried_data) == 0:
            raise IndexError("Query returned no data!")

        timestamps = np.fromiter([datum.timestamp for datum in queried_data], dtype=float)
        values = np.fromiter([datum.value for datum in queried_data], dtype=float)

        # The start and end times of the returned data may not be the same as what was requested,
        # and we need the actual values to store the metadata properly
        actual_start_time = datetime.fromtimestamp(timestamps[0], tz=timezone.utc)
        actual_end_time = datetime.fromtimestamp(timestamps[-1], tz=timezone.utc)

        # Subtract off the first timestamp so every timestamp is delta-time from the first.
        timestamps -= timestamps[0]

        # Reform the x-axis as a homogenous array
        x_axis = np.arange(timestamps[0], timestamps[-1], granularity)

        # Interpolate the data array onto this new x-axis
        wave = np.interp(x_axis, timestamps, values)

        time_series_dict = {
            "start": actual_start_time,
            "stop": actual_end_time,
            "car": "N/A",
            "measurement": "N/A",
            "field": field,
            "granularity": granularity,
            "length": len(x_axis),
            "units": get_data_units(get_sensor_id(field)),
        }

        return TimeSeries(wave, time_series_dict)

    def init_schema(self):
        """
        Initialize the Postgres database schema. This only needs to be called ONCE PER DATABASE.

        Has no effect if called on an already initialized database.
        """
        init_schema(self._engine)

    def get_session(self) -> Session:
        """
        Obtain a new Session to make queries to this database.
        :return: a new Session bound to this connection
        """
        return self._session_builder()

    def write(self, instance: Union[object, List[object]], session: Session = None) -> None:
        """
        Write some data ``instance`` to this database using ``session``.

        :param instance: the data that will be uploaded, can be a single object or an iterable of objects.
        :param session: optionally specify the Session that will be used to write to the database
        :raises RuntimeError: if the database write fails
        """
        # Use the default session if none was specified
        if session is None:
            session = self._session

        try:
            # Upload data as a batch if it's a list
            if isinstance(instance, List):
                session.add_all(instance)
            else:
                session.add(instance)

            session.commit()

        except Exception as e:
            # If the database write failed, ensure to roll back the commit
            session.rollback()
            raise RuntimeError("Failed to write to PostgreSQL!") from e

        
if __name__ == "__main__":
    field = "TotalPackVoltage"
    client = PostgresClient()

    start_time = datetime(2024, 6, 16, 10, 0, 0, tzinfo=timezone.utc)
    end_time = datetime(2024, 8, 21, 23, 0, 0, tzinfo=timezone.utc)

    data: TimeSeries = client.query(field, start_time, end_time, granularity=0.1)

    x_axis = data.datetime_x_axis
    plt.plot(x_axis, data)
    plt.show()



from typing import Callable
from data_tools.schema import Result, CanonicalPath, File


class FileLoader:
    """
    A callable object that wraps a lambda function to acquire a `File`, as well as the
    canonical path that will be queried for the `File` in question.
    """
    def __init__(self, loader: Callable[[CanonicalPath], Result[File]], canonical_path: CanonicalPath) -> None:
        """
        Wrap a `loader` along with the `canonical_path` that it will be querying.

        :param loader: A single argument lambda expecting a canonical path that will return
            either a `File` wrapped in a `Result` or raise a `FileNotFoundError`.
        :param canonical_path: The path that the lambda will be querying.
        """
        self._loader: Callable[[CanonicalPath], Result[File]] = loader
        self._canonical_path: CanonicalPath = canonical_path

    def __call__(self) -> Result[File]:
        """
        Invoke this `FileLoader`, and obtain a `Result` containing a File` or an `FileNotFoundError` if it cannot be found.

        :raises FileNotFoundError: If the `File` cannot be loaded.
        :return: The `File` that was loaded
        """
        return self._loader(self._canonical_path)

    @property
    def canonical_path(self) -> CanonicalPath:
        return self._canonical_path



from sqlalchemy.orm import declarative_base
from sqlalchemy import Column, Integer, Float, UniqueConstraint, Engine

Base = declarative_base()

targets = ["TotalPackVoltage", "BatteryCurrent", "BatteryVoltage", "VehicleVelocity"]

_data_id_encodings = {
    "TotalPackVoltage": 0,
    "BatteryCurrent": 1,
    "BatteryVoltage": 2,
    "VehicleVelocity": 3,
    "PackCurrent": 4,
    "BatteryCurrentDirection": 5,
    "VoltageofLeast": 6
}


_data_units = {
    0: "V",
    1: "A",
    2: "V",
    3: "m/s",
    4: "A",
    5: "A.U.",
    6: "V"
}


def get_sensor_id(data_field: str) -> int:
    """
    Return the canonical data ID corresponding to the data titled ``data_field``.

    :param data_field: the data's name as a string
    :return: the data's canonical ID as an integer
    """
    return _data_id_encodings[data_field]


def get_data_units(data_id: int) -> str:
    """
    Return the units of the data corresponding to the canonical ID ``data_id``.

    :param int data_id: the data's canonical ID as an integer
    :return: the data's units as a string
    """
    return _data_units[data_id]


class CANLog(Base):
    """

    Encapsulates a single CAN log message within an SQL database.

    """
    __tablename__ = "sensor_data"

    id = Column(Integer, primary_key=True)          # The unique ID of this entry in the database
    timestamp = Column(Float, nullable=False)       # The timestamp as a UTC POSIX timestamp
    sensor_type = Column(Integer, nullable=False)   # The canonical data ID of this log
    value = Column(Float, nullable=False)           # The value of this measurement

    # Enforce that there should be no entries of the same data at the same time!
    __table_args__ = (
        UniqueConstraint('timestamp', 'sensor_type', name='unique_timestamp_sensor_type'),
    )


def init_schema(engine: Engine):
    """
    Initialize this schema in the database connected to by ``engine``.

    :param engine: connection to database that will be mutated with this schema
    """
    Base.metadata.create_all(engine)



import base64
from dotenv import load_dotenv
import os
import requests
from data_tools.schema import File, Result, CanonicalPath
import dill
import json


load_dotenv()


class SunbeamCache:
    def __init__(self, api_url: str = None):
        """
        Create a client to connect to the Sunbeam Cache API.

        Uses ``api_url`` if set, or by default "api.sunbeam.ubcsolar.com/cache".
        """
        self._base_url = api_url or "api.sunbeam.ubcsolar.com/cache"

    def _build_url(self, *components):
        """
        Construct the full URL for the given endpoint components.
        """
        return "http://" + "/".join([self._base_url] + list(components))

    @staticmethod
    def _handle_error(response):
        """Handle HTTP errors consistently."""
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError:
            reason = response.reason
            if isinstance(reason, bytes):
                reason = reason.decode("utf-8")
            raise RuntimeError(reason)

    def __getitem__(self, item):
        url = self._build_url("get")
        response = requests.get(url, params={"key": item})

        # If we found the cached item, return it
        if response.status_code == 200:
            encoded_data = response.content
            serialized_data = base64.b64decode(encoded_data)

            return dill.loads(serialized_data)

        # Otherwise if the item wasn't found
        elif response.status_code == 406:
            raise KeyError(response.text)

        # Or if we encountered some other error
        else:
            SunbeamCache._handle_error(response)

    def __setitem__(self, key, value):
        url = self._build_url("set")
        serialized_data = dill.dumps(value, protocol=dill.HIGHEST_PROTOCOL)
        encoded_data = base64.b64encode(serialized_data).decode("utf-8")
        response = requests.get(url, params={"key": key, "value": encoded_data})

        # Handle the error if the set failed
        if response.status_code != 201:
            SunbeamCache._handle_error(response)

    def __contains__(self, item):
        url = self._build_url("exists")
        response = requests.get(url, params={"key": item})

        # Decode the response to a bool, as Flask encodes `True` into `"True"` when emitting the response
        if response.status_code == 200:
            return response.text.strip().lower() == "true"

        else:
            SunbeamCache._handle_error(response)

    def __delitem__(self, key):
        url = self._build_url("delete")
        response = requests.get(url, params={"key": key})

        if response.status_code != 200:
            SunbeamCache._handle_error(response)

    def keys(self):
        url = self._build_url("keys")
        response = requests.get(url)

        if response.status_code != 200:
            SunbeamCache._handle_error(response)

        else:
            return json.loads(response.content)


class SunbeamClient:
    """

    Encapsulate a client connection to the Sunbeam API, UBC Solar's custom data pipeline.

    """
    def __init__(self, api_url: str = None):
        """
        Create a client to connect to the Sunbeam API.

        Uses the `SUNBEAM_URL` environment variable if ``api_url`` is not set, or by default "api.sunbeam.ubcsolar.com".
        """
        if api_url is None:
            api_url = os.getenv("SUNBEAM_URL") if "SUNBEAM_URL" in os.environ.keys() else "api.sunbeam.ubcsolar.com"

        self._base_url = api_url

    @property
    def cache(self):
        """
        Acquire an interface to the Sunbeam Cache API, descended from this client.
        """
        return SunbeamCache(self._base_url + "/cache")

    def distinct(self, key: str, filter: dict[str, str]) -> list[str]:
        """
        Obtain the distinct values of a given key, subject to a filter.

        For example, if you want to see the distinct events processed by a certain pipeline,
        you can set `key` to `"event"` and `filter` to `{"origin": "your_pipeline"}`.

        As another example, if you wanted to see the events processed by a certain pipeline where the "power" stage
        was processed, you can set `key` to `"event"` and `filter` to `{"origin": "your_pipeline", "source": "power"}`.

        :param key: name of the field for which we want to get the distinct values
        :param filter: a mapping which filters the query to have items match this filter's values by field
        :return: the values corresponding the `key` field of matching items
        """
        url_components: list[str] = [self._base_url, "files", "distinct"]
        url = "http://" + "/".join(url_components)
        data = {"key": key}
        data.update(filter)

        response = requests.post(url, data=data)

        if response.status_code == 200:
            return json.loads(response.text)

        # Otherwise, acquire and wrap the error
        else:
            # Try to extract the error and capture it
            response.raise_for_status()



    def get_file(self, origin: str = None, event: str = None, source: str = None, name: str = None, path: CanonicalPath = None) -> Result[File | Exception]:
        """
        Get a File from the Sunbeam API.

        You must provide either a ``CanonicalPath`` object directory by setting ``path``, OR provide ALL the
        individual path elements ``origin``, ``event``, ``source``, and ``name``.

        :return: a ``Result`` wrapping a ``File`` or an ``Exception``
        :raises AssertionError: if ``path`` was not provided and any of ``origin``, ``event``, ``source``, ``name`` are None.
        """
        # Start the url as api.sunbeam.ubcsolar.com/files
        url_components: list[str] = [self._base_url, "files"]

        # Prefer using ``path`` to build the URL
        if path is not None:
            url_components.extend(path.unwrap())

        else:
            # Make sure if we are using the individual path elements, all of them were provided
            if None in [origin, event, source, name]:
                raise AssertionError("All of ``origin``, ``event``, ``source``, and ``name`` cannot be none!")

            url_components.extend([origin, event, source, name])

        # Build the path, and set file_type=bin to request the binary data of the File we want
        url = "http://" + "/".join(url_components)
        params = {'file_type': "bin"}

        response = requests.get(url, params=params)

        # If we got the File successfully, wrap and return the deserialized File object
        if response.status_code == 200:
            serialized_data = response.content
            return Result.Ok(dill.loads(serialized_data))

        # Otherwise, acquire and wrap the error
        else:
            # Try to extract the error and capture it
            try:
                response.raise_for_status()

            except requests.exceptions.HTTPError as e:
                return Result.Err(e)

            # We didn't capture an error, but we still didn't get an OK error code, so wrap
            # the response message in a RuntimeError
            if isinstance(response.reason, bytes):
                reason = response.reason.decode("utf-8")
            else:
                reason = response.reason

            return Result.Err(RuntimeError(reason))



from data_tools.lap_tools.fsgp_2024_laps import FSGPDayLaps
from data_tools.query.influxdb_query import DBClient
from typing import Callable
import numpy as np

def collect_lap_data(query_func: Callable, client: DBClient, include_day_2=False,
                     verbose=False) -> np.ndarray:
    """
    Higher order function - computes `query_func` for each lap in FSGP 2024 and returns the resulting array.

    Set `include_day_2` to True to include the day 2 laps, which were driven slowly & under heavy rain.

    Example usage:

    ```python
    from data_tools.collections import collect_lap_data, TimeSeries
    from data_tools.query import DBClient
    import datetime
    import numpy as np


    def get_average_speed(start_time: datetime.datetime, end_time: datetime.datetime, data_client: DBClient):
        lap_speed: TimeSeries = data_client.query_time_series(start_time, end_time, "VehicleVelocity")
        return np.mean(lap_speed)


    client = DBClient()

    average_speeds = collect_lap_data(get_average_speed, client)
    ```

    :param Callable query_func: must take in parameters (lap_start: datetime, lap_end:datetime, data_client:DBClient)
    :param DBClient client: client to use for querying
    :param include_day_2: flag to include the three day 2 laps, driven slowly & under heavy rain
    :param verbose: if True, print out queried data during execution
    :return: a NumPy ndarray of `query_func` results for all laps
    """

    if include_day_2:
        laps = (FSGPDayLaps(1),  # Corresponds to July 16th
                FSGPDayLaps(2),  # Corresponds to July 17th
                FSGPDayLaps(3))  # Corresponds to July 18th
    else:
        laps = (FSGPDayLaps(1),
                FSGPDayLaps(3))

    indices = [range(day_laps.get_lap_count()) for day_laps in laps]

    lap_data = []
    # Iterate through all selected laps
    for day_laps, lap_indices in zip(laps, indices):
        for lap_idx in lap_indices:
            lap_num = lap_idx + 1
            lap_start = day_laps.get_start_utc(lap_num)
            lap_end = day_laps.get_finish_utc(lap_num)
            lap_data.append(query_func(lap_start, lap_end, client))
            if verbose:
                print(f"Processed data for day {day_laps.day} lap {lap_num}")
                print(f"{lap_start=}\n{lap_end=}")
                print(f"{query_func.__name__} result for lap {lap_num}: {lap_data[-1]}\n")

    return np.array(lap_data)



"""
===========================================
Collections (:mod:`data_tools.collections`)
===========================================

Time Series Data
===========
.. autosummary::
   :toctree: generated/

   TimeSeries    -- Enhanced `ndarray` with powerful data analysis features

"""
from .time_series import TimeSeries


__all__ = [
    "TimeSeries",
]



from .query import (
    FluxQuery,
    FluxStatement,
    DBClient,
    PostgresClient,
    SunbeamClient
)

from .collections import (
    TimeSeries,
)

from .schema import (
    FileType,
    File,
    Event,
    DataSource,
    FileLoader
)

from .utils import (
    parse_iso_datetime,
    ensure_utc,
    iso_string_from_datetime
)

from .lap_tools import (
    FSGPDayLaps,
    collect_lap_data
)


__all__ = [
    "FluxQuery",
    "FluxStatement",
    "TimeSeries",
    "DBClient",
    "FSGPDayLaps",
    "FSGPDayLaps",
    "collect_lap_data",
    "PostgresClient",
    "parse_iso_datetime",
    "ensure_utc",
    "iso_string_from_datetime",
    "FileType",
    "File",
    "Event",
    "DataSource",
    "FileLoader",
    "SunbeamClient"
]
